#include <stdio.h>
#include <math.h>
#include "iodefine.h"
#include "bldc.h"

/* Reset-synchronized PWM module.  All cycles are active in the late
   part of the cycle, so they end right at the interrupt.  */

static double cosine_table[] = {
#include "cosines.h"
};

/* These numbers give us a cycle time of exactly 20 KHz (50 uS) */
#define CYCLE 4000
#define MAXPOWER (CYCLE)

static void __attribute__((interrupt))
cpwm_handler ()
{
  static int bit = 0;
  MTU4.TSR.BIT.TCFV = 0;
  bit += 0x10;
  PORTB.DDR.BYTE = 0xf0;
  PORTB.DR.BYTE = bit;
}

void
cpwm_init()
{
  
  /* [1] Clear the CST3 and CST4 bits in TSTRA to 0 to stop the TCNT
         operation. Specify the reset-synchronized PWM mode while
         MTU3.TCNT and MTU4.TCNT are stopped. */

  MTU.TSTRA.BIT.CST3 = 0;
  MTU.TSTRA.BIT.CST4 = 0;

  /* [2] Set bits TPSC[2:0] and CKEG[1:0] in MTU3.TCR to select the
	 counter clock and clock edge for channel 3. Set bits
	 CCLR[2:0] to select TGRA compare match as a counter clear
	 source. */

  MTU3.TCR.BIT.TPSC = 0;	/* ICLK/1 */
  MTU3.TCR.BIT.CKEG = 0;
  MTU3.TCR.BIT.CCLR = 1;	/* TGRA */

  MTU4.TCR.BIT.TPSC = 0;	/* ICLK/1 */
  MTU4.TCR.BIT.CKEG = 0;
  MTU4.TCR.BIT.CCLR = 1;	/* TGRA */

  MTU3.TIORH.BIT.IOB = 1;	/* 15.17 MTIOC3B */
  MTU3.TIORL.BIT.IOD = 1;	/* 15.30 MTIOC3D */
  MTU4.TIORH.BIT.IOA = 1;	/* 15.31 MTIOC4A*/
  MTU4.TIORL.BIT.IOC = 1;	/* 15.32 MTIOC4C*/
  MTU4.TIORH.BIT.IOB = 1;	/* 15.19 MTIOC4B */
  MTU4.TIORL.BIT.IOD = 1;	/* 15.20 MTIOC4D */

  /* [3] When performing brushless DC motor control, set bit BDC in
         the timer gate control register (TGCRA) and set the feedback
         signal input source and output chopping or gate signal direct
         output. */

  MTU.TGCRA.BIT.BDC = 0;
  MTU.TGCRA.BIT.P = 1;
  MTU.TGCRA.BIT.N = 1;
  MTU.TGCRA.BIT.FB = 0;

  /* [4] Set MTU3.TCNT and MTU4.TCNT to 0000h. */

  MTU3.TCNT = 0;
  MTU4.TCNT = 0;

  /* [5] MTU3.TGRA is the cycle register. Set the waveform cycle value
         in MTU3.TGRA. Set the transition timing of the PWM output
         waveforms in MTU3.TGRB, MTU4.TGRA, and MTU4.TGRB. Set times
         within the compare match range of MTU3.TCNT. */

  MTU3.TGRA = CYCLE;

  MTU3.TGRB = CYCLE*1/6;
  MTU4.TGRA = CYCLE*2/6;
  MTU4.TGRB = CYCLE*3/6;

  /* [6] Enable or disable toggle output synchronized with the PMW
         cycle using bit PSYE in TOCR1A, and set the PWM output level
         with bits OLSP and OLSN. When specifying the PWM output level
         by using TOLBRA as a buffer for TOCR2A, see figure 15.3. */

  MTU.TOCR1A.BIT.TOCL = 0;
  MTU.TOCR1A.BIT.PSYE = 1; /* this conflicts with ENCA */
  MTU.TOCR1A.BIT.OLSP = 1;
  MTU.TOCR1A.BIT.OLSN = 1;
  MTU.TOCR1A.BIT.TOCS = 0;

  MTU.TOCR2A.BYTE = 0x3f;
  MTU.TOCR2B.BYTE = 0x3f;

  MTU.TOLBRA.BYTE = 0xff;
  MTU.TOLBRB.BYTE = 0xff;

  /* [7] Set bits MD[3:0] in MTU3.TMDR1 to 1000b to select the
         reset synchronized PWM mode. In addition, specify the buffer
         operation of the MTU3.TGRA (BFA = 1) and MTU3.TGRC (BFB = 1)
         when necessary. Set the MTU4.TMDR1 to the initial value
         (normal operation; BFA = 0 and BFB = 0). */

  MTU4.TMDR.BYTE = 0x00;
  MTU3.TMDR.BYTE = 0x08;

  /* [8] Set the enabling/disabling of the PWM waveform output
     pin in TOERA. */

  MTU.TOERA.BYTE = 0x3f;
  
  /* [12] Set the port control register and the port I/O register. */

  PORT7.DDR.BYTE = UH | UL | VH | VL | WH | WL;

  /* [9] Set the CST3 bit in the TSTRA to 1 to start the count
     operation. */

  MTU.TSTRA.BYTE |= 0xc0;

#if 1
  /* pwm cycle interrupt */
  _vectors[VECT(MTU4,TCIV4)] = cpwm_handler;
  IEN(MTU4,TCIV4) = 1;
  IPR(MTU4,TCIV4) = 9;
  MTU4.TIER.BIT.TCIEV = 1;
#endif

  /* Current sense ADC */
  S12AD1.ADANS.WORD = 0x2000;
  S12AD1.ADCER.BIT.ADRFMT = 1;
  S12AD1.ADCSR.BYTE = 0x4c;
  S12AD1.ADCSR.BYTE = 0xcc;
}

static int
power2level (int x)
{
  int rv;
  /* Turn -32768..32767 into MAXPOWER..0 */
  /*x *= 1.414;*/
  rv = MAXPOWER/2 - x * (MAXPOWER/2) / 32768;
  if (rv < 0)
    rv = 0;
  if (rv > CYCLE)
    rv = CYCLE;
  return rv;
}

char pwm_status[200];
int pwm_status_new = 0;
static int saved_torque = 0;
static int last_min=-1, cur_min=-1;
static unsigned short motor_current_peak = 0;
static int bits = 0;

float motor_current = 0;

void
cpwm_levels (int u, int v, int w)
{
  unsigned int pu, pv, pw, pmax, motor_current_i;
  int adj;
  fflush (stdout);
  pu = power2level (u);
  pv = power2level (v);
  pw = power2level (w);

#if 0
  if (pu > pv && pu > pw)
    {
      pmax = pu;
      cur_min = UL;
    }
  else if (pv > pw)
    {
      pmax = pv;
      cur_min = VL;
    }
  else
    {
      cur_min = WL;
      pmax = pw;
    }
  adj = (CYCLE) - pmax;
  pu += adj;
  pv += adj;
  pw += adj;
  if (pu >= CYCLE/2) pu = CYCLE;
  if (pv >= CYCLE/2) pv = CYCLE;
  if (pw >= CYCLE/2) pw = CYCLE;
#endif

  MTU3.TGRB = pu;
  MTU4.TGRA = pv;
  MTU4.TGRB = pw;
  iprintf("U %6d  V %6d  W %6d\n", pu, pv, pw);

  switch (cur_min)
    {
      bits = bits ^ 3;
    case UL:
      motor_current_i = UCURR;
      break;
    case VL:
      motor_current_i = VCURR;
      break;
    case WL:
      motor_current_i = WCURR;
      break;
    }
  if (motor_current_peak < motor_current_i)
    motor_current_peak = motor_current_i;

  if (cur_min != last_min)
    {
      motor_current = ((int)motor_current_peak - 32768) / 22282.24;
      motor_current_peak = 0;
      last_min = cur_min;
    }
#if 0
  sprintf(pwm_status, "\033[5;1H\033[%dmPWM %6d %6d %6d = max %04x adj %04x %04x:%04x:%04x\033[0m\n",
	  (saved_torque > 0) ? 32 : (saved_torque < 0) ? 31 : 0, u, v, w, pmax, adj, pu, pv, pw);
  pwm_status_new = 1;
#endif
}

static double
dcos (int pos)
{
  pos *= 2;
  pos %= COS_TABLE_SIZE*4;
  if (pos < 0)
    pos += COS_TABLE_SIZE*4;
  switch (pos / COS_TABLE_SIZE) {
  case 0:
    return cosine_table[pos];
  case 1:
    return - cosine_table[COS_TABLE_SIZE*2 - pos];
  case 2:
    return - cosine_table[pos - COS_TABLE_SIZE*2];
  case 3:
    return cosine_table[COS_TABLE_SIZE*4 - pos];
  }
}

static int old_pos = 8000;

int cpwm_step;

void
cpwm_torque (int torque)
{
  int pos = encoder_ang (); /* 0..1999 */


  if (torque == 0)
    {
      MTU3.TGRD = 
	MTU4.TGRC = 
	MTU4.TGRD = CYCLE;
      sprintf(pwm_status, "\033[5;1HPWM off\033[K\n");
      pwm_status_new = 1;
      return;
    }

  if (abs (pos - old_pos) < 300)
    pos += (pos-old_pos) / 2;

  saved_torque = torque;
  cpwm_levels (torque * dcos(pos - ULEAD - LEAD),
	       torque * dcos(pos - VLEAD - LEAD),
	       torque * dcos(pos - WLEAD - LEAD));
  saved_torque = 0;

  cpwm_step = pos - old_pos;
  old_pos = pos;
}

void
bldc_run ()
{
  MTU.TOERA.BYTE = 0x3f;
}

void
bldc_stop ()
{
  MTU.TOERA.BYTE = 0x00;
  PORT7.DR.BYTE = 0;
  while (encoder_rpm () > 500)
    ;
  PORT7.DR.BYTE = UL | VL | WL;
}
